from abc import ABC, abstractmethod
from dataclasses import dataclass
from importlib.metadata import version
from pathlib import Path
from typing import Generator

from mountaineer.client_builder.parser import ControllerWrapper
from mountaineer.paths import ManagedViewPath


@dataclass
class ParsedController:
    """Represents a fully parsed controller with its associated paths and metadata"""

    wrapper: ControllerWrapper
    view_path: ManagedViewPath
    url_prefix: str | None = None
    is_layout: bool = False


class FileGeneratorBase(ABC):
    def __init__(self, *, managed_path: Path):
        self.managed_path = managed_path

        mountaineer_version = version("mountaineer")
        self.standard_header = CodeBlock(
            "/*",
            f" * This file was generated by Mountaineer v{mountaineer_version}. Do not edit it manually.",
            " */",
        )

    def build(self):
        blocks = list(self.script())
        blocks = [self.standard_header] + blocks
        self.managed_path.write_text("\n\n".join(block.content for block in blocks))

    @abstractmethod
    def script(self) -> Generator["CodeBlock", None, None]:
        pass


class CodeBlock:
    """
    Semantic grouping of a particular section of code, typically separated
    from other ones with two blank lines.

    """

    def __init__(self, *lines: str):
        self.lines = lines

    @classmethod
    def indent(cls, line: str):
        """
        Use the first line of the code block to determine the indentation level
        for the other lines that are implicitly imbedded in this line. This most
        acutely covers cases where string templating includes newlines so we can
        preserve the overall layout.

        ie. CodeBlock.indent(f"  my_var = {my_var}\n")

        """
        if not line or "\n" not in line:
            return line

        has_trailing_newline = line.endswith("\n")
        lines = line[:-1].split("\n") if has_trailing_newline else line.split("\n")

        # Get the base indentation from first line
        base_indent, _ = cls._get_indent_level(lines[0])

        # Process all lines
        result: list[str] = []
        result.append(lines[0])  # First line remains unchanged

        # Add base indentation to subsequent lines while preserving their own
        for current_line in lines[1:]:
            if current_line.strip():  # If line has content
                result.append(base_indent + current_line)
            else:  # Preserve empty or whitespace-only lines as-is
                result.append(current_line)

        return "\n".join(result) + ("\n" if has_trailing_newline else "")

    @classmethod
    def _get_indent_level(cls, line: str) -> tuple[str, int]:
        """
        Get the indentation string and count from the start of a line.
        Returns tuple of (indent_str, indent_count).
        """
        indent_str = ""
        for char in line:
            if char not in (" ", "\t"):
                break
            indent_str += char
        return indent_str, len(indent_str)

    @property
    def content(self):
        return "\n".join(self.lines)



from importlib.metadata import version
from pathlib import Path
from re import match as re_match

import questionary
from click import command, option, secho

from create_mountaineer_app.builder import build_project
from create_mountaineer_app.enums import PackageManager
from create_mountaineer_app.environments.poetry import PoetryEnvironment
from create_mountaineer_app.environments.uv import UvEnvironment
from create_mountaineer_app.external import (
    get_git_user_info,
)
from create_mountaineer_app.generation import EditorType, ProjectMetadata


def prompt_package_manager() -> PackageManager:
    """
    Prompt the user to choose a package manager and handle its installation if needed.
    """
    package_manager = questionary.select(
        "Choose a package manager:",
        choices=[
            PackageManager.UV,
            PackageManager.POETRY,
            PackageManager.VENV,
        ],
        default=PackageManager.UV,
    ).unsafe_ask()

    if package_manager == PackageManager.POETRY:
        # Helpful utilities to wrap poetry logic and lifecycle
        poetry_environment = PoetryEnvironment()

        # Determine if the user already has poetry
        if not poetry_environment.has_provider():
            input_install_poetry = questionary.confirm(
                "Poetry is not installed. Install poetry?", default=True
            ).unsafe_ask()
            if input_install_poetry:
                secho("Installing poetry...")
                try:
                    poetry_environment.install_provider()
                except Exception as e:
                    secho(f"Error installing poetry: {e}", fg="red")
                    raise e

                input_add_to_path = questionary.confirm(
                    "Add poetry to PATH? [Yes]", default=True
                ).unsafe_ask()
                if input_add_to_path:
                    poetry_environment.add_poetry_to_path()

    elif package_manager == PackageManager.UV:
        uv_environment = UvEnvironment()

        # Determine if the user already has uv
        if not uv_environment.has_provider():
            input_install_uv = questionary.confirm(
                "uv is not installed. Install uv?", default=True
            ).unsafe_ask()
            if input_install_uv:
                secho("Installing uv...")
                try:
                    uv_environment.install_provider()
                except Exception as e:
                    secho(f"Error installing uv: {e}", fg="red")
                    raise e

    return package_manager


def prompt_author() -> tuple[str, str]:
    while True:
        git_name, git_email = get_git_user_info()
        default_author = (
            f"{git_name} <{git_email}>" if git_name and git_email else "Name <email>"
        )
        author_combined = (
            questionary.text(f"Author [{default_author}]").unsafe_ask()
            or default_author
        )

        # Parse the author into a name and email
        matched_obj = re_match(r"(.*) <(.*)>", author_combined)

        if not matched_obj:
            secho(
                f"Author must be in the format 'Name <email>', not '{author_combined}'",
                fg="red",
            )
            continue

        return matched_obj.group(1), matched_obj.group(2)


def get_current_version_number():
    """
    Return the version of the mountaineer app that is linked to this particular
    template project layout.

    We assume the version numbers are tied, which they are deterministically in CI.

    """
    return version("create_mountaineer_app")


@command()
@option("--output-path", help="The output path for the bundled files.")
@option("--mountaineer-dev-path", help="The path to the Mountaineer dev environment.")
def main(output_path: str | None, mountaineer_dev_path: str | None):
    """
    Create a new Mountaineer project.

    Most of the configuration is handled interactively. We have a few CLI arguments for rarer options that can configure the installation behavior.

    :param output_path: The output path for the bundled files, if not the current directory.

    """
    input_project_name = (
        questionary.text("Project name [my-project]:").unsafe_ask() or "my-project"
    )
    input_author_name, input_author_email = prompt_author()
    package_manager = prompt_package_manager()
    input_create_stub_files = questionary.confirm(
        "Create stub MVC files? [Yes]", default=True
    ).unsafe_ask()
    input_use_tailwind = questionary.confirm(
        "Use Tailwind CSS? [Yes]", default=True
    ).unsafe_ask()
    input_editor_config = questionary.rawselect(
        "Add editor configuration? [vscode]",
        choices=["vscode", "vim", "zed", "no"],
        default="vscode",
    ).unsafe_ask()

    secho("\nCreating project...", fg="green")

    project_path = Path(output_path) if output_path else Path.cwd() / input_project_name
    project_path = project_path.resolve()

    metadata = ProjectMetadata(
        project_name=input_project_name.replace(" ", "_").replace("-", "_"),
        author_name=input_author_name,
        author_email=input_author_email,
        package_manager=package_manager,
        use_tailwind=input_use_tailwind,
        editor_config=(
            EditorType.from_name(input_editor_config)
            if input_editor_config != "no"
            else None
        ),
        project_path=project_path,
        create_stub_files=input_create_stub_files,
        mountaineer_min_version=get_current_version_number(),
        mountaineer_dev_path=(
            Path(mountaineer_dev_path).resolve() if mountaineer_dev_path else None
        ),
    )

    build_project(metadata)


if __name__ == "__main__":
    main()



import traceback
from dataclasses import dataclass, field
from pathlib import Path
from time import time

from firehot import Environment
from firehot.environment import IsolatedProcess

from mountaineer.console import CONSOLE
from mountaineer.development.isolation import IsolatedAppContext
from mountaineer.development.messages import (
    BootupMessage,
    BuildJsMessage,
    BuildUseServerMessage,
    StartServerMessage,
)
from mountaineer.development.messages_broker import (
    AsyncMessageBroker,
    BrokerServerConfig,
)
from mountaineer.io import async_to_sync
from mountaineer.logging import LOGGER


@dataclass
class FileChangesState:
    pending_js: set[Path] = field(default_factory=set)
    pending_python: set[Path] = field(default_factory=set)

    current_context: IsolatedProcess | None = None


@dataclass
class WebserverConfig:
    host: str
    port: int
    live_reload_port: int


@dataclass
class IsolatedContext:
    webcontroller: str
    webserver_config: WebserverConfig | None
    message_config: BrokerServerConfig


@async_to_sync
async def run_isolated(
    isolated_context: IsolatedContext,
):
    """
    Isolated subprocess with all runtime modules already injected by firehot.

    """
    app_context = IsolatedAppContext.from_webcontroller(
        webcontroller=isolated_context.webcontroller,
    )

    try:
        async with AsyncMessageBroker.new_client(
            isolated_context.message_config
        ) as broker:
            await app_context.run_async(broker)
    except Exception as e:
        # This logging should happen automatically
        CONSOLE.print(f"[red]Error: {e}")
        CONSOLE.print(traceback.format_exc())
        raise e


async def restart_backend(
    environment: Environment,
    broker: AsyncMessageBroker,
    state: FileChangesState,
    isolated_context: IsolatedContext,
):
    CONSOLE.print("Restarting backend")
    if state.current_context is not None:
        # Drain any pending messages from the previous process
        LOGGER.debug("Draining message queue")
        jobs = await broker.drain_queue()
        if jobs:
            LOGGER.debug(f"Dropped {len(jobs)} pending messages")
        environment.stop_isolated(state.current_context)

    # We might have updated imports, so pass to the env to optionally update
    # No-op if no dependencies have changed, so the subsequent exec should be instantaneous
    environment.update_environment()

    state.current_context = environment.exec(
        run_isolated,
        isolated_context,
    )

    # Bootstrap the process and rebuild the server files
    CONSOLE.print("[bold cyan]🚀 Booting server...[/bold cyan]")
    start = time()
    await broker.send_and_get_response(BootupMessage())

    if isolated_context.webserver_config is not None:
        await broker.send_and_get_response(
            StartServerMessage(
                host=isolated_context.webserver_config.host,
                port=isolated_context.webserver_config.port,
                live_reload_port=isolated_context.webserver_config.live_reload_port,
            )
        )

    CONSOLE.print("[bold yellow]⚙️  Building useServer components...[/bold yellow]")
    await broker.send_and_get_response(BuildUseServerMessage())
    build_time = time() - start
    CONSOLE.print(
        f"[bold green]✨ Backend build complete in {build_time:.2f}s![/bold green]"
    )


async def rebuild_frontend(broker: AsyncMessageBroker, state: FileChangesState):
    start = time()
    CONSOLE.print("[bold yellow]🔨 Rebuilding frontend...[/bold yellow]")

    await broker.send_and_get_response(
        # None will rebuild everything - we want this in cases where we are called
        # without a list provided
        BuildJsMessage(updated_js=list(state.pending_js) if state.pending_js else None)
    )

    build_time = time() - start
    CONSOLE.print(
        f"[bold green]✨ Frontend build complete in {build_time:.2f}s![/bold green]"
    )



from dataclasses import dataclass
from pathlib import Path
from typing import Generic, TypeVar

TResponse = TypeVar("TResponse")


@dataclass
class IsolatedMessageBase(Generic[TResponse]):
    """Base class for all messages passed between main process and isolated app context"""

    pass


@dataclass
class ErrorResponse:
    """Generic error response"""

    exception: str
    traceback: str


@dataclass
class SuccessResponse:
    """Generic success response"""

    pass


@dataclass
class BootupMessage(IsolatedMessageBase[SuccessResponse | ErrorResponse]):
    """Message to bootup the isolated app context"""

    pass


@dataclass
class StartServerMessage(IsolatedMessageBase[SuccessResponse | ErrorResponse]):
    """Message to start the server"""

    host: str
    port: int
    live_reload_port: int


@dataclass
class BuildJsMessage(IsolatedMessageBase[SuccessResponse | ErrorResponse]):
    """Message to trigger JS compilation"""

    updated_js: list[Path] | None


@dataclass
class BuildUseServerMessage(IsolatedMessageBase[SuccessResponse | ErrorResponse]):
    """Message to build the useServer support files"""

    pass


MessageTypes = (
    BootupMessage | StartServerMessage | BuildJsMessage | BuildUseServerMessage
)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Type, TypeAlias

from mountaineer.client_compiler.base import APIBuilderBase
from mountaineer.controller import ControllerBase
from mountaineer.controller_layout import LayoutControllerBase


@dataclass
class BuildConfig:
    custom_builders: list[APIBuilderBase] = field(default_factory=list)


# Proper type alias definition
CONTROLLER_TYPE: TypeAlias = ControllerBase | LayoutControllerBase


@dataclass
class MountaineerPlugin:
    name: str
    controllers: list[Type[CONTROLLER_TYPE]]

    view_root: Path

    build_config: BuildConfig

    _concrete_controllers: dict[Type[CONTROLLER_TYPE], CONTROLLER_TYPE] = field(
        default_factory=dict
    )

    def init_controller(self, controller: CONTROLLER_TYPE):
        """
        User-override if they need to specifically initialize a controller with
        custom logic.

        """
        if type(controller) in self._concrete_controllers:
            raise ValueError(f"Controller {type(controller)} already initialized")

        self._concrete_controllers[type(controller)] = controller

    def get_controllers(self):
        """
        All clients that use this plugin should call this function to get the
        concrete controllers.

        """
        # Initialize if not already initialized
        for controller_cls in self.controllers:
            if controller_cls not in self._concrete_controllers:
                # Assume these controllers have no init params
                self._concrete_controllers[controller_cls] = controller_cls()

        return list(self._concrete_controllers.values())

    def to_webserver(self):
        """
        Shortcut utility to create a webserver from the plugin to allow
        for easy builds of the output files.

        """
        from mountaineer.app import AppController

        app_controller = AppController(
            view_root=self.view_root,
            custom_builders=self.build_config.custom_builders,
        )
        for controller in self.controllers:
            app_controller.register(controller())
        return app_controller



"""
Utilities to generate some common Typescript objects from
Python definitions.

"""

from typing import Any

from inflection import camelize


class TSLiteral(str):
    """
    This string should not be quoted when used in a Typescript insert.
    """

    def __add__(self, other):
        # Concatenate self (TSLiteral) with other (any string)
        # and return a new TSLiteral object
        if isinstance(other, str):
            return TSLiteral(super(TSLiteral, self).__add__(other))
        return NotImplemented

    def __radd__(self, other):
        # Handle right-side addition, where a string is added to a TSLiteral
        if isinstance(other, str):
            return TSLiteral(other.__add__(self))
        return NotImplemented


def python_payload_to_typescript(
    payload: Any, dict_equality: str = ":", current_indent: int = 0
) -> str:
    """
    Take an element with python tokens that should be outputted to
    Typescript and consolidate them into a valid payload.

    """
    indent_str = " " * current_indent
    inner_indent_str = " " * (current_indent + 2)

    if isinstance(payload, dict):
        # Convert the children
        children_lines: list[str] = []
        for key, value in payload.items():
            if isinstance(value, TSLiteral):
                # If the literal is the same as the key, we don't need to include it since it's referencing
                # the same variable.
                if key == value:
                    children_lines.append(f"{inner_indent_str}{key}")
                    continue

            key = python_payload_to_typescript(key)
            value = python_payload_to_typescript(
                value, current_indent=current_indent + 2
            )
            children_lines.append(f"{inner_indent_str}{key}{dict_equality} {value}")

        children_str = ",\n".join(children_lines)
        return "\n".join(
            [
                "{",
                children_str,
                f"{indent_str}}}",
            ]
        )
    elif isinstance(payload, list):
        children_lines = [python_payload_to_typescript(child) for child in payload]
        children_str = ",\n".join(
            [f"{inner_indent_str}{child}" for child in children_lines]
        )
        return f"[\n{children_str}\n{indent_str}]"
    elif isinstance(payload, TSLiteral):
        return payload
    elif isinstance(payload, str):
        return f"'{payload}'"
    elif isinstance(payload, bool):
        return str(payload).lower()
    elif isinstance(payload, (int, float)):
        return str(payload)
    elif payload is None:
        return "null"
    else:
        raise ValueError(
            f"Unknown payload type {type(payload)} for Typescript conversion."
        )


def normalize_interface(title: str):
    replace_chars = {" ", "[", "]"}

    for char in replace_chars:
        title = title.strip(char)
        title = title.replace(char, "_")

    return camelize(title)



from typing import Any, Type, cast, get_type_hints

from fastapi import HTTPException
from pydantic import BaseModel, Field, create_model

from mountaineer.annotation_helpers import MountaineerUnsetValue

# Keys that are used to initialize the HTTPException model
HTTPExceptionKeys = ["status_code", "detail", "headers"]


class APIExceptionInternalModelBase(BaseModel):
    """
    Superclass used for our synthetic internal errors. This class sets
    the required parameters and default model configuration for use in
    validating user inputs to APIException(**kwargs)

    """

    status_code: int
    detail: str
    headers: dict[str, str]

    model_config = {"extra": "forbid"}


class InternalModelMeta(type):
    """
    Introspect APIException class definitions where they're defined to convert
    their class-based typehints into an internal Pydantic BaseModel that can validate
    individual instances.

    """

    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        cls._create_internal_model()
        return cls

    def _create_internal_model(cls):
        type_hints = get_type_hints(cls)

        fields = {
            key: (
                key_type,
                getattr(cls, key, cls._build_default_field(key, key_type)),
            )
            for key, key_type in type_hints.items()
            if key not in ["InternalModel", "internal_model"]
        }
        cls.InternalModel = create_model(
            # Mirror the class name so our OpenAPI objects are as the user specifies
            # for their exception class
            cls.__name__,
            __base__=APIExceptionInternalModelBase,
            **cast(Any, fields),
        )
        cls.InternalModel.__module__ = cls.__module__

    def _build_default_field(cls, key, key_type):
        default_value = getattr(cls, key, MountaineerUnsetValue())

        if isinstance(default_value, MountaineerUnsetValue):
            return Field()
        else:
            return Field(default_factory=lambda: default_value)

    def __call__(cls, *args, **kwargs):
        # Override the __call__ method to instantiate models like Pydantic does
        # Use the internal model for validation and instantiation
        internal_model = cls.InternalModel(**kwargs)
        instance = super().__call__(
            **{
                key: value
                for key, value in internal_model.model_dump().items()
                if key in HTTPExceptionKeys
            }
        )
        setattr(instance, "internal_model", internal_model)
        return instance


class APIException(HTTPException, metaclass=InternalModelMeta):
    """
    Base class for user defined APIExceptions that can be thrown
    in server actions and should provide some metadata back to
    the client caller.

    ```python
    class PostNotFound(APIException):
        status_code: int = 404
        detail: str = "The post was not found"
        post_id: int
        is_deleted: bool

    class MyController(ControllerBase):
        @passthrough
        def get_post(self, post_id: int) -> Post:
            post = self.post_service.get_post(post_id)
            if not post:
                raise PostNotFound(post_id=post_id, is_deleted=True)
            return post
    ```

    """

    status_code: int = 500
    detail: str = "A server error occurred"
    headers: dict[str, str] = Field(default_factory=dict)

    # Set by the metaclass to provide internal validation for runtime values assigned
    # to our marked up typehints
    InternalModel: Type[APIExceptionInternalModelBase]

    # Set on the instance of the exception with the user values, these are
    # used to pass to the client caller
    internal_model: APIExceptionInternalModelBase

    # We can't synthetically create an API contract with the instance variables
    # (pydantic uses a plugin to get their IDE typehints)
    # Any issues with the input constructor will raise a runtime error
    # versus being statically checked
    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)


class RequestValidationFailure(BaseModel):
    error_type: str
    location: list[str]
    message: str
    value_input: Any


class RequestValidationError(APIException):
    """
    Exception to be raised when a Pydantic model or url parameters fails to validate.

    """

    status_code: int = 422
    detail: str = "Request validation failed"
    errors: list[RequestValidationFailure]



import asyncio
import os
import signal
from os import environ
from pathlib import Path
from random import uniform
from shutil import copytree
from subprocess import Popen
from time import sleep, time

import httpx
import pytest
import toml

from mountaineer.__tests__.fixtures import get_fixture_path
from mountaineer.cli import (
    find_packages_with_prefix,
)


@pytest.fixture
def tmp_ci_webapp(tmp_path: Path):
    # Copy the full ci_webapp package so we can make local modifications
    # just within this test
    raw_package = get_fixture_path("ci_webapp")
    mutable_package = tmp_path / "ci_webapp"
    copytree(raw_package, mutable_package)

    pyproject_path = mutable_package / "pyproject.toml"
    base_package_path = (get_fixture_path("") / "../../../").resolve()

    with open(pyproject_path, "r") as file:
        content = toml.load(file)

    # Point to the absolute path of the local mountaineer core package, versus the
    # symlinked version in the original package. We only have one dependency so we can
    # just replace the entire bundle.
    assert len(content["project"]["dependencies"]) == 1
    content["project"]["dependencies"] = [
        f"mountaineer @ file://{str(base_package_path)}"
    ]

    with open(pyproject_path, "w") as file:
        toml.dump(content, file)

    return mutable_package


def test_find_packages_with_prefix():
    # Choose some packages that we know will be in the test environment
    assert set(find_packages_with_prefix("fasta")) == {"fastapi"}
    assert set(find_packages_with_prefix("pydan")) == {
        "pydantic",
        "pydantic_core",
        "pydantic-settings",
    }


async def check_server_bound(port: int, timeout=8):
    # 5s hard timeout + 3s overhead
    # When the server restarting gets stuck it gets stuck permanently
    start_time = time()
    url = f"http://localhost:{port}"
    async with httpx.AsyncClient() as client:
        while time() - start_time < timeout:
            try:
                response = await client.get(url)
                return True, response.status_code
            except httpx.RequestError:
                pass
            await asyncio.sleep(0.1)
    return False, -1


@pytest.mark.integration_tests
@pytest.mark.asyncio
async def test_handle_runserver_with_user_modifications(tmp_ci_webapp: Path):
    # Ensure that there is no existing webapp running
    port = 5006
    url = f"http://localhost:{port}"
    async with httpx.AsyncClient() as client:
        try:
            await client.get(url, timeout=1)
            assert False, "The server is already running"
        except httpx.RequestError:
            pass

    uv_env = {
        key: value
        for key, value in environ.items()
        if not key.startswith("VIRTUAL_ENV")
    }

    # We need to poetry install the packages at the new path
    return_code = Popen(["uv", "sync"], cwd=tmp_ci_webapp, env=uv_env).wait()
    assert return_code == 0

    return_code = Popen(
        ["npm", "install"], cwd=tmp_ci_webapp / "ci_webapp" / "views", env=uv_env
    ).wait()
    assert return_code == 0

    # Start the handle_runserver function in a process
    server_process = Popen(
        ["uv", "run", "runserver", "--port", str(port)],
        cwd=tmp_ci_webapp,
        env=uv_env,
    )
    test_file_path = tmp_ci_webapp / "ci_webapp" / "controllers" / "home.py"

    try:
        for _ in range(5):
            with open(test_file_path, "a") as f:
                print(f"Adding content to {test_file_path}")  # noqa: T201
                f.write("\npass\n")

            sleep(uniform(0.2, 2.0))

        # After all these random server restarts make sure that the
        # server is still running
        print(  # noqa: T201
            "Done with changes, checking that server will resolve if not immediately ready..."
        )
        is_bound, status_code = await check_server_bound(port)
        assert is_bound, "Server is not bound to localhost:3000"
        assert status_code == 200, "Server is not returning 200 status code"
        print("Server is bound to expected port")  # noqa: T201
    finally:
        # Terminate the processes after test
        os.kill(server_process.pid, signal.SIGKILL)
        server_process.wait()



from copy import copy
from dataclasses import dataclass
from enum import Enum
from inspect import isclass
from typing import (
    Any,
    Callable,
    Generator,
    Optional,
    Type,
    TypeVar,
    Union,
)

from fastapi import APIRouter
from fastapi.params import Body, Depends, File, Form, Header
from fastapi.routing import APIRoute
from inflection import camelize
from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo

from mountaineer.actions.fields import (
    FunctionActionType,
    FunctionMetadata,
    get_function_metadata,
)
from mountaineer.client_builder.types import TypeDefinition, TypeParser
from mountaineer.constants import STREAM_EVENT_TYPE
from mountaineer.controller import (
    ControllerBase,
    class_fn_as_method,
    get_client_functions_cls,
)
from mountaineer.controller_layout import LayoutControllerBase
from mountaineer.exceptions import APIException as APIException
from mountaineer.render import RenderBase

T = TypeVar("T")


class WrapperName:
    # The original name given to the object, just used for record-keeping.
    raw_name: str

    # Must be globally unique.
    global_name: str

    # The name used in the local context as a shortcut while working with a single controller
    local_name: str

    def __init__(self, name: str):
        self.raw_name = name
        self.local_name = name
        self.global_name = name


@dataclass
class CoreWrapper:
    name: WrapperName
    module_name: str


@dataclass
class FieldWrapper:
    name: str
    value: Union["ModelWrapper", "EnumWrapper", "TypeDefinition", type]
    required: bool


@dataclass
class ModelWrapper(CoreWrapper):
    model: type[BaseModel]
    isolated_model: type[BaseModel]  # Model with only direct fields
    superclasses: list["ModelWrapper"]
    value_models: list[FieldWrapper]
    body_type: str = "application/json"


@dataclass
class ExceptionWrapper(CoreWrapper):
    status_code: int
    exception: Type[APIException]
    value_models: list[FieldWrapper]


@dataclass
class EnumWrapper(CoreWrapper):
    enum: type[Enum]


@dataclass
class ActionWrapper:
    name: str
    module_name: str
    action_type: FunctionActionType

    params: list[FieldWrapper]
    headers: list[FieldWrapper]
    request_body: Optional[ModelWrapper]
    response_bodies: dict[Type[ControllerBase], ModelWrapper | None]
    exceptions: list[ExceptionWrapper]

    is_raw_response: bool
    is_streaming_response: bool

    # Actions can be mounted to multiple controllers through inheritance
    # This will store a mapping of each controller to the url that the action is mounted to
    controller_to_url: dict[Type[ControllerBase], str]

    def has_required_params(self):
        return (
            (any([param.required for param in self.params]) if self.params else False)
            or (
                any([header.required for header in self.headers])
                if self.headers
                else False
            )
            or self.request_body is not None
        )


@dataclass
class ControllerWrapper(CoreWrapper):
    entrypoint_url: str | None
    controller: type[ControllerBase]
    superclasses: list["ControllerWrapper"]

    # Render entrypoint
    queries: list[FieldWrapper]
    paths: list[FieldWrapper]
    render: Optional[ModelWrapper]

    # Actions
    actions: dict[
        str, ActionWrapper
    ]  # {url: action} directly implemented for this controller

    @property
    def all_actions(self) -> list[ActionWrapper]:
        # Convert each action. We also include the superclass methods, since they're
        # actually bound to the controller instance with separate urls.
        all_actions: list[ActionWrapper] = []

        # If an action is overridden in a subclass, we shouldn't include it twice
        # Unlike other traversal functions, we'd rather identify these actions
        # by their names and not memory values because we want to only show the lowest
        # subclassed implementation
        seen_actions: set[str] = set()

        def parse_controller(controller):
            for action in controller.actions.values():
                if action.name in seen_actions:
                    continue
                all_actions.append(action)
                seen_actions.add(action.name)
            for superclass in controller.superclasses:
                parse_controller(superclass)

        parse_controller(self)
        return all_actions

    @classmethod
    def get_all_embedded_types(
        cls, controllers: list["ControllerWrapper"], include_superclasses: bool = False
    ) -> "EmbeddedTypeContainer":
        """
        For all the models and enums that are embedded in this controller (actions+render), return them in a flat list.
        Results will be deduplicated.

        :param include_superclasses: If provided, we will also traverse up the hierarchy to include all models
        referenced by superclasses.

        """
        models: list[ModelWrapper] = []
        enums: list[EnumWrapper] = []
        exceptions: list[ExceptionWrapper] = []

        def _traverse_logic(
            item: ControllerWrapper
            | ModelWrapper
            | ExceptionWrapper
            | ActionWrapper
            | EnumWrapper
            | TypeDefinition,
        ):
            nonlocal models, enums

            if isinstance(item, ControllerWrapper):
                if item.render:
                    yield item.render

                yield from item.actions.values()

                if include_superclasses:
                    yield from item.superclasses

            elif isinstance(item, ActionWrapper):
                if item.request_body:
                    yield item.request_body
                if item.response_bodies:
                    yield from item.response_bodies.values()
                yield from item.exceptions

            elif isinstance(item, ModelWrapper):
                models.append(item)

                for field in item.value_models:
                    yield field.value

                if include_superclasses:
                    yield from item.superclasses

            elif isinstance(item, EnumWrapper):
                enums.append(item)

            elif isinstance(item, ExceptionWrapper):
                exceptions.append(item)

                for field in item.value_models:
                    yield field.value

            elif isinstance(item, TypeDefinition):
                yield from item.children

        cls._traverse_iterator(_traverse_logic, controllers)
        return EmbeddedTypeContainer(models=models, enums=enums, exceptions=exceptions)

    @classmethod
    def get_all_embedded_controllers(
        cls, controllers: list["ControllerWrapper"]
    ) -> list["ControllerWrapper"]:
        """
        Gets all unique superclasses of the given set of controllers.

        """
        all_controllers: list[ControllerWrapper] = []

        def _traverse_logic(item: ControllerWrapper):
            nonlocal all_controllers

            all_controllers.append(item)
            yield from item.superclasses

        cls._traverse_iterator(_traverse_logic, controllers)
        return all_controllers

    @classmethod
    def _traverse_iterator(
        cls,
        logic: Callable[[T | Any], Generator[T | Any, None, None]],
        initial_queue: list[T],
    ):
        """
        Memory-identity traversal, only will traverse each unique object once.
        Clients write a logic function that returns the next items to traverse, and meanwhile
        can make use of our single-traversal guarantee to store results in a flat list.

        ex:

        ```
        models = []
        def logic(item):
            if isinstance(item, ControllerWrapper):
                if item.render:
                    yield item.render
                for superclass in item.superclasses:
                    yield superclass
            elif isinstance(item, ModelWrapper):
                models.append(item)

        ```

        """
        queue = copy(initial_queue)
        already_seen: set[int] = set()
        while queue:
            item = queue.pop(0)
            if id(item) in already_seen:
                continue
            queue.extend(list(logic(item)))
            already_seen.add(id(item))


@dataclass
class SelfReference:
    name: str
    model: Type[BaseModel]


@dataclass
class EmbeddedTypeContainer:
    models: list[ModelWrapper]
    enums: list[EnumWrapper]
    exceptions: list[ExceptionWrapper]


class ControllerParser:
    """
    Our ControllerParser is responsible for taking the in-memory representations of
    ControllerBase models and extracting the metadata required to convert them
    through the TypeScript pipeline into full interface signatures and implementations.

    """

    def __init__(self):
        self.parsed_models: dict[Type[BaseModel], ModelWrapper] = {}
        self.parsed_enums: dict[Type[Enum], EnumWrapper] = {}
        self.parsed_controllers: dict[Type[ControllerBase], ControllerWrapper] = {}
        self.parsed_self_references: list[SelfReference] = []
        self.parsed_exceptions: dict[Type[APIException], ExceptionWrapper] = {}

        self.type_parser = TypeParser()

    def parse_controller(self, controller: type[ControllerBase]) -> ControllerWrapper:
        """Main entry point to parse a controller into intermediary representation"""
        if controller in self.parsed_controllers:
            return self.parsed_controllers[controller]

        # Get all valid superclasses in MRO order. Include any controller that is either explicitly a subclass
        # of ControllerBase or has client functions defined on it.
        controller_classes = self._get_valid_parent_classes(
            controller,
            base_require_predicate=lambda base: (
                issubclass(base, ControllerBase)
                or len(list(get_client_functions_cls(base))) > 0
            ),
            base_exclude_classes=(ControllerBase, LayoutControllerBase),
        )

        # Get render model from the concrete controller
        render, render_path, render_query, entrypoint_url = self._parse_render(
            controller
        )
        actions = self._parse_actions(controller)

        # Parse superclasses
        superclass_controllers: list[ControllerWrapper] = []
        for superclass in controller_classes:
            superclass_controllers.append(self.parse_controller(superclass))

        wrapper = ControllerWrapper(
            name=WrapperName(controller.__name__),
            module_name=controller.__module__,
            entrypoint_url=entrypoint_url,
            controller=controller,
            actions=actions,
            queries=render_query or [],
            paths=render_path or [],
            render=render,
            superclasses=superclass_controllers,
        )
        self.parsed_controllers[controller] = wrapper
        return wrapper

    def _parse_model(
        self, model: type[BaseModel], skip_object_ids: tuple | None = None
    ) -> ModelWrapper:
        """Parse a Pydantic model into ModelWrapper, handling inheritance"""
        # Return cached if already parsed
        if model in self.parsed_models:
            return self.parsed_models[model]

        # Get all valid superclasses in MRO order, excluding BaseModel and above
        model_classes = self._get_valid_parent_classes(
            model, base_require=BaseModel, base_exclude_classes=(BaseModel, RenderBase)
        )

        # Parse direct superclasses (excluding the model itself)
        superclasses: list[ModelWrapper] = []
        for base in model_classes:
            if base not in self.parsed_models:
                # Now parse it properly
                self.parsed_models[base] = self._parse_model(base)
            superclasses.append(self.parsed_models[base])

        # Handle fields, excluding those from superclasses
        fields: list[FieldWrapper] = []
        isolated_model = self._create_isolated_model(model)
        for name, field in isolated_model.model_fields.items():
            if field.exclude:
                continue
            # No user schema will self-reference the isolated model, it will only
            # reference the original definition
            fields.append(self._parse_field(name, field, self_model=model))

        wrapper = ModelWrapper(
            name=WrapperName(model.__name__),
            module_name=model.__module__,
            model=model,
            isolated_model=isolated_model,
            superclasses=superclasses,
            value_models=fields,
        )
        self.parsed_models[model] = wrapper
        return wrapper

    def _parse_field(
        self,
        name: str,
        field_info: FieldInfo,
        self_model: Type[BaseModel] | None = None,
    ) -> FieldWrapper:
        # Create a basic conversion of the types, in case they're wrapped
        # by complex types like List, Dict, etc.
        root_definition = self.type_parser.parse_type(field_info.annotation)

        # Now we can recursively parse the children. We want to wrap all of the sub-models
        # in their own wrapper objects.
        def update_children(type_definition: TypeDefinition | type):
            if isinstance(type_definition, TypeDefinition):
                type_definition.update_children(
                    [update_children(child) for child in type_definition.children]
                )
                return type_definition
            else:
                # Special case to avoid infinite recursion
                if self_model and type_definition == self_model:
                    reference = SelfReference(
                        name=self_model.__name__, model=self_model
                    )
                    self.parsed_self_references.append(reference)
                    return reference

                # Determine if they qualify for conversion. The vast majority of values
                # passed in here will be classes, since they represent the typehinted annotations
                # of models. But there are some situations (like TypeVars used in generics) where
                # they will fail a subclass check.
                if isclass(type_definition) and issubclass(type_definition, BaseModel):
                    return self._parse_model(type_definition)
                elif isclass(type_definition) and issubclass(type_definition, Enum):
                    return self._parse_enum(type_definition)
                else:
                    # No need to parse further
                    return type_definition

        root_definition = update_children(root_definition)

        return FieldWrapper(
            name=name,
            value=root_definition,  # type: ignore
            required=field_info.is_required(),
        )

    def _parse_enum(self, enum_type: type[Enum]) -> EnumWrapper:
        """Parse an Enum into EnumWrapper"""
        if enum_type in self.parsed_enums:
            return self.parsed_enums[enum_type]

        wrapper = EnumWrapper(
            name=WrapperName(enum_type.__name__),
            module_name=enum_type.__module__,
            enum=enum_type,
        )
        self.parsed_enums[enum_type] = wrapper
        return wrapper

    def _parse_render(
        self, controller: type[ControllerBase]
    ) -> tuple[
        ModelWrapper | None,
        list[FieldWrapper] | None,
        list[FieldWrapper] | None,
        str | None,
    ]:
        """Parse the render method's return type"""
        render = getattr(controller, "render", None)
        if not render:
            return None, None, None, None

        try:
            metadata = get_function_metadata(render)
        except AttributeError:
            return None, None, None, None

        return_model = metadata.get_render_model()
        if not return_model:
            return None, None, None, None

        # Only standard controllers will have url mounts. For layout controllers since they don't
        # mount to an actual path in the router it's fine to use any synthetic path.
        # This also applies to inherited render methods that come from a parent, they
        # can only use query params.
        entrypoint_url = metadata.controller_mounts.get(controller)

        # Only parse models and params for concrete render() implementation, not parent classes
        # that just inherit the ControllerBase's ABC generic signature
        model_schema = self._parse_model(return_model)
        path_params, query_params = self._parse_params(
            class_fn_as_method(render), "render", entrypoint_url or "/render"
        )

        return model_schema, path_params, query_params, entrypoint_url

    def _create_isolated_model(
        self,
        model: type[BaseModel],
    ) -> type[BaseModel]:
        """
        Create a new model with only the direct fields (no inherited fields).
        Handles both regular Pydantic models and generic model instances.

        """
        generic_origin: Type[BaseModel] | None = None
        generic_args: tuple[Any, ...] | None = None

        # For generic models, we need to synthesize annotations from the generic metadata
        if hasattr(model, "__pydantic_generic_metadata__"):
            generic_metadata = model.__pydantic_generic_metadata__
            generic_origin = generic_metadata["origin"]
            generic_args = generic_metadata["args"]

        if generic_origin and generic_args:
            # Build annotations dict by resolving generic types just for the fields that
            # were defined directly on the superclass. Since we're iterating with model_fields
            # on the synthetically created generic subclass, all of the annotations should be resolved
            # to real types by this point
            parent_owned_fields = generic_origin.__dict__.get("__annotations__", {})

            return create_model(  # type: ignore
                model.__name__,
                __config__=model.model_config,
                **{
                    # Since we're modifying the annotation types, we need to copy the full
                    # field_info since .annotation will be set on the new model. Without a copy
                    # it will affect the original model's state.
                    field_name: (field_info.annotation, copy(field_info))  # type: ignore
                    for field_name, field_info in model.model_fields.items()
                    if field_name in parent_owned_fields
                },
            )
        else:
            # Regular model - use original logic
            include_fields = {
                field_name: (field_info.annotation, field_info)
                for field_name, field_info in model.model_fields.items()
                if field_name in model.__dict__.get("__annotations__", {})
            }
            return create_model(  # type: ignore
                model.__name__,
                __config__=model.model_config,
                **include_fields,  # type: ignore
            )

    def _create_temp_route(self, func: Callable, name: str, url: str) -> APIRoute:
        """Create a temporary FastAPI route using the actual function"""
        router = APIRouter()
        router.add_api_route(
            # We need to use the right path so it can separate out the path paramss
            # from the query params
            path=f"/{url}",
            endpoint=func,
            # We don't use the FastAPI's sniffed response model parsing in our pipeline, and
            # some definitions like server-side streaming AsyncIterables can't be handled
            # natively by FastAPI
            response_model=None,
        )

        route = next(
            route
            for route in router.routes
            if isinstance(route, APIRoute) and route.path == f"/{url}"
        )

        return route

    def _parse_params(self, func: Callable, name: str, url: str):
        """Parse route parameters using FastAPI's dependency system"""
        route = self._create_temp_route(func, name, url)
        path_params: list[FieldWrapper] = [
            self._parse_field(
                name=param.name,
                field_info=param.field_info,
            )
            for param in route.dependant.path_params
        ]
        query_params: list[FieldWrapper] = [
            self._parse_field(
                name=param.name,
                field_info=param.field_info,
            )
            for param in route.dependant.query_params
            if not isinstance(param.field_info, (Body, Header, Depends))
        ]

        return path_params, query_params

    def _parse_headers(self, func: Callable, name: str, url: str) -> list[FieldWrapper]:
        """Parse header parameters using FastAPI's dependency system"""
        route = self._create_temp_route(func, name, url)
        headers: list[FieldWrapper] = []

        for param in route.dependant.header_params:
            field = self._parse_field(
                name=param.name,
                field_info=param.field_info,
            )
            headers.append(field)

        return headers

    def _parse_request_body(
        self, func: Callable, name: str, url: str
    ) -> Optional[ModelWrapper]:
        """Parse request body using FastAPI's dependency system"""
        route = self._create_temp_route(func, name, url)

        if not route.dependant.body_params:
            return None

        body_fields: dict[str, FieldInfo] = {}
        has_files = False
        has_form = False

        # Analyze all body parameters
        for body_param in route.dependant.body_params:
            field_info = body_param.field_info
            field_type = body_param.type_

            # Handle Pydantic models
            if isinstance(field_type, type) and issubclass(field_type, BaseModel):
                return self._parse_model(field_type)

            # Handle File uploads, subclass of Form
            elif isinstance(field_info, File):
                has_files = True
                body_fields[body_param.name] = field_info

            # Handle Form fields
            elif isinstance(field_info, Form):
                has_form = True
                body_fields[body_param.name] = field_info

        # Create dynamic model for the body, since forms don't otherwise
        # have a Pydantic model that wraps the values
        body_model = create_model(
            f"{camelize(name)}Form",
            __module__=func.__module__,
            **{
                name: (field_info.annotation, field_info)  # type: ignore
                for name, field_info in body_fields.items()
            },
        )

        # Determine the type of request body
        if has_files:
            body_type = "multipart/form-data"
        elif has_form:
            body_type = "application/x-www-form-urlencoded"
        else:
            return None

        model_wrapped = self._parse_model(body_model)
        model_wrapped.body_type = body_type
        return model_wrapped

    def _parse_response_bodies(self, metadata: FunctionMetadata):
        """Parse response model from metadata"""
        return {
            controller: self._parse_model(model)
            if issubclass(model, BaseModel)
            else None
            for controller, model in metadata.return_models.items()
        }

    def _parse_actions(self, controller: type) -> dict[str, ActionWrapper]:
        """Parse all actions in a controller"""
        actions: dict[str, ActionWrapper] = {}

        for name, func, metadata in get_client_functions_cls(controller):
            # We don't need a url for the action, since actions can't take path
            # parameters all kwargs will just be query params
            synthetic_action_url = f"/{name}"

            path_params, query_params = self._parse_params(
                func, name, synthetic_action_url
            )
            action = ActionWrapper(
                name=name,
                module_name=controller.__module__,
                params=query_params,
                headers=self._parse_headers(func, name, synthetic_action_url),
                request_body=self._parse_request_body(func, name, synthetic_action_url),
                response_bodies=self._parse_response_bodies(metadata),
                is_raw_response=metadata.is_raw_response,
                is_streaming_response=metadata.media_type == STREAM_EVENT_TYPE,
                exceptions=[
                    self._parse_exception(exception)
                    for exception in metadata.exception_models
                ],
                action_type=metadata.action_type,
                controller_to_url=metadata.controller_mounts,
            )
            actions[name] = action

        return actions

    def _parse_exception(self, exception: Type[APIException]):
        if exception in self.parsed_exceptions:
            return self.parsed_exceptions[exception]

        value_models = [
            self._parse_field(name, field_info)
            for name, field_info in exception.InternalModel.model_fields.items()
        ]

        # Unlike standard Models, which are used 1:1 to validate client bodies where some
        # fields are required and others can safely be left missing to default to a
        # server-side value, we know exceptions will be fully populated at runtime. The
        # non-required fields within InternalModel only indicate at exception instance
        # creation what clients have to supply for every constructor versus what
        # is inherited from the superclass.
        for field in value_models:
            field.required = True

        wrapper = ExceptionWrapper(
            name=WrapperName(exception.__name__),
            module_name=exception.__module__,
            status_code=exception.status_code,
            exception=exception,
            value_models=value_models,
        )
        self.parsed_exceptions[exception] = wrapper
        return wrapper

    def _get_valid_parent_classes(
        self,
        cls: type,
        *,
        base_require: type | None = None,
        base_require_predicate: Callable[[type], bool] | None = None,
        base_exclude_classes: tuple[type, ...],
    ) -> list[type]:
        """
        Helper to get valid MRO parents, excluding certain base classes

        """
        return [
            base
            for base in cls.__bases__
            if (
                base not in base_exclude_classes
                and base is not object
                and (not base_require or issubclass(base, base_require))
                and (not base_require_predicate or base_require_predicate(base))
            )
        ]



from pathlib import Path

from mountaineer import AppController, ConfigBase, ControllerBase


class TestController(ControllerBase):
    view_path = "/test_controller/page.tsx"
    url = "/"

    async def render(self) -> None:
        pass


# We need a development config to test dev utilities
class SimpleConfig(ConfigBase):
    pass


config = SimpleConfig()

test_controller = AppController(
    view_root=Path(__file__).parent.joinpath("views"),
    config=config,
)
test_controller.register(TestController())



import socket


def get_free_port() -> int:
    """
    Leverage the OS-port shortcut :0 to get a free port. Return the value
    of the port that was assigned.

    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        port = s.getsockname()[1]
        s.close()
    return port



import importlib
import os
from importlib.metadata import distributions
from pathlib import Path


def find_packages_with_prefix(prefix: str):
    """
    Find and return a list of all installed package names that start with the given prefix.

    """
    return [
        dist.metadata["Name"]
        for dist in distributions()
        if dist.metadata["Name"].startswith(prefix)
    ]


def package_path_to_module(package: str, file_path_raw: Path) -> str:
    """
    Convert a file path to its corresponding Python module path.

    Args:
        package: The root package name (e.g. 'amplify')
        file_path_raw: The file path to convert

    Returns:
        The full module path (e.g. 'amplify.controllers.auth')
    """
    # Get the package's root directory
    package_module = importlib.import_module(package)
    if not package_module.__file__:
        raise ValueError(f"The package {package} does not have a __file__ attribute")

    package_root = os.path.dirname(package_module.__file__)
    file_path = os.path.abspath(str(file_path_raw))

    # Check if the file is within the package
    if not file_path.startswith(package_root):
        raise ValueError(
            f"The file {file_path} is not in the package {package} ({package_root})"
        )

    # Remove the package root and the file extension
    relative_path = os.path.relpath(file_path, package_root)
    module_path = os.path.splitext(relative_path)[0]

    # Convert path separators to dots and add the package name
    return f"{package}.{module_path.replace(os.sep, '.')}"



from uuid import UUID, uuid4

from fastapi import Request
from mountaineer import ControllerBase, Metadata, RenderBase


class ComplexRender(RenderBase):
    client_ip: str
    random_uuid: UUID
    delay_loops: int
    throw_client_error: bool


class ComplexController(ControllerBase):
    """
    We set up our Complex controller and view to simulate what a long running
    React render looks on the client side. Because of our naive prime-number loop,
    each "delay_loop" is expected to saturate the process with work for about 2s.

    Adusting the `delay_loops` parameter on the view will allow you to set
    the number of loops and therefore the total rendering time that it takes.
    This in-turn lets you test our timeout of 10s for a rendering job.

    """

    url = "/complex/{detail_id}/"
    view_path = "/app/complex/page.tsx"

    def __init__(self):
        super().__init__(
            hard_ssr_timeout=5,
        )

    def render(
        self,
        detail_id: UUID,
        request: Request,
        delay_loops: int | None = None,
        throw_client_error: bool = False,
    ) -> ComplexRender:
        return ComplexRender(
            client_ip=request.client.host if request.client else "unknown",
            random_uuid=uuid4(),
            metadata=Metadata(title=f"Complex: {detail_id}"),
            delay_loops=delay_loops or 0,
            throw_client_error=throw_client_error,
        )



from pathlib import Path

from mountaineer.app import AppController
from mountaineer.client_compiler.compile import ClientCompiler


def test_build_static_metadata(tmpdir: Path):
    app = AppController(view_root=tmpdir)
    compiler = ClientCompiler(app=app)

    # Write test files to the view path to determine if we're able
    # to parse the whole file tree
    static_dir = compiler.view_root.get_managed_static_dir()

    (static_dir / "test_css.css").write_text("CSS_TEXT")

    (static_dir / "nested").mkdir(exist_ok=True)
    (static_dir / "nested" / "test_nested.css").write_text("CSS_TEXT")

    # File contents are the same - shas should be the same as well
    metadata = compiler._build_static_metadata()
    assert "test_css.css" in metadata.static_artifact_shas
    assert "nested/test_nested.css" in metadata.static_artifact_shas
    assert (
        metadata.static_artifact_shas["test_css.css"]
        == metadata.static_artifact_shas["nested/test_nested.css"]
    )



from pydantic_settings import BaseSettings

from mountaineer.config import ConfigBase


class Settings1(BaseSettings):
    ABC: str


class Settings2(BaseSettings):
    DEF: str


class Config(ConfigBase, Settings1, Settings2):
    pass


def test_merge_configs():
    """
    Ensure that we can merge two configurations together, both
    during type checking and during runtime.

    """
    config = Config(ABC="abc", DEF="def")
    assert config.ABC == "abc"
    assert config.DEF == "def"



import pytest

from mountaineer.io import lru_cache_async


@pytest.mark.asyncio
async def test_lru_cache_async_infinite_cache():
    exec_counts = 0

    @lru_cache_async(maxsize=None)
    async def cache_values():
        nonlocal exec_counts
        exec_counts += 1
        return exec_counts

    assert await cache_values() == 1
    assert await cache_values() == 1
    assert await cache_values() == 1


@pytest.mark.asyncio
async def test_lru_cache_async_limited_cache():
    exec_counts = 0

    @lru_cache_async(maxsize=2)
    async def cache_values(a: int):
        nonlocal exec_counts
        exec_counts += 1
        return exec_counts

    assert await cache_values(1) == 1
    assert await cache_values(1) == 1
    assert await cache_values(1) == 1

    assert await cache_values(2) == 2

    # At this point the cache is full with [1, 2]
    # If the next value is out of scope, it will clear
    # the oldest value (1)
    assert await cache_values(3) == 3

    # This new call will recompute the value for 1
    assert await cache_values(1) == 4



import asyncio
import socket
from functools import lru_cache, wraps
from typing import Any, Callable, Coroutine, ParamSpec, TypeVar


async def gather_with_concurrency(
    tasks: list[Coroutine[Any, Any, Any]],
    *,
    n: int,
    catch_exceptions: bool = False,
) -> list[Any]:
    semaphore = asyncio.Semaphore(n)

    async def sem_task(task: Coroutine[Any, Any, Any]) -> Any:
        async with semaphore:
            return await task

    return await asyncio.gather(
        *(sem_task(task) for task in tasks),
        return_exceptions=catch_exceptions,
    )


T = TypeVar("T")
P = ParamSpec("P")


def async_to_sync(async_fn: Callable[P, Coroutine[Any, Any, T]]) -> Callable[P, T]:
    @wraps(async_fn)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(async_fn(*args, **kwargs))
        return result

    return wrapper


def get_free_port() -> int:
    """
    Leverage the OS-port shortcut :0 to get a free port. Return the value
    of the port that was assigned.

    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        port = s.getsockname()[1]
        s.close()
    return int(port)


def lru_cache_async(
    maxsize: int | None = 100,
):
    def decorator(
        async_function: Callable[..., Coroutine[Any, Any, T]],
    ):
        @lru_cache(maxsize=maxsize)
        @wraps(async_function)
        def internal(*args, **kwargs):
            coroutine = async_function(*args, **kwargs)
            # Unlike regular coroutine functions, futures can be awaited multiple times
            # so our caller functions can await the same future on multiple cache hits
            return asyncio.ensure_future(coroutine)

        return internal

    return decorator
